"""The middleware for scrapy-wayback."""
import http
import typing

import scrapy
import wayback

from .response import WaybackMachineResponse


class WaybackMachineDownloaderMiddleware:

    @classmethod
    def from_crawler(cls, crawler):
        s = cls()
        crawler.signals.connect(s.spider_opened, signal=scrapy.signals.spider_opened)
        return s

    def process_request(self, request, spider):
        wayback_proxy_enabled = request.meta.get("wayback_machine_proxy_enabled", spider.settings.get("WAYBACK_MACHINE_PROXY_ENABLED", False))
        wayback_proxy_fallthrough_enabled = request.meta.get("wayback_machine_proxy_fallthrough_enabled", spider.settings.get("WAYBACK_MACHINE_PROXY_FALLTHROUGH_ENABLED", True))
        if wayback_proxy_enabled and request.method == "GET":
            wayback_response = WaybackMachineResponse(request, self.client.search(request.url), None, self.client)
            if wayback_response.is_valid():
                return wayback_response
        if not wayback_proxy_fallthrough_enabled:
            return None
        return request

    def process_response(self, request, response, spider):
        fallback_enabled = request.meta.get("wayback_machine_fallback_enabled", spider.settings.get("WAYBACK_MACHINE_FALLBACK_ENABLED", True))
        if not fallback_enabled:
            return response
        if response.status >= http.HTTPStatus.BAD_REQUEST and request.method == "GET":
            wayback_response = WaybackMachineResponse(request, self.client.search(request.url), response, self.client)
            if wayback_response.is_valid():
                return wayback_response
        return response

    def spider_opened(self, spider):
        self.client = wayback.WaybackClient()



"""The standard wayback response."""
import typing
import http

import scrapy
import wayback


def find_memento(
    records: typing.List[str],
    client: wayback.WaybackClient
) -> typing.Tuple[typing.Optional[wayback.Memento], typing.Optional[str]]:
    """Finds the latest valid memento in the records."""
    for record in records:
        try:
            memento = client.get_memento(record, mode=wayback.Mode.original)
            if memento.status_code == http.HTTPStatus.OK:
                return memento, record
        except wayback.exceptions.MementoPlaybackError:
            pass
    return None, None


class WaybackMachineResponse(scrapy.http.HtmlResponse):
    """A response object containing information about the wayback response."""
    def __init__(
        self,
        request: scrapy.Request,
        records: typing.List[str],
        response: typing.Optional[scrapy.http.Response],
        client: wayback.WaybackClient
    ) -> None:
        memento, record = find_memento(records, client)
        self._memento = memento
        self._record = record
        self._records = records
        self._client = client
        super().__init__(
            request.url,
            body=self._memento.text,
            encoding=self._memento.encoding,
            status=self._memento.status_code,
            request=request,
        )
        self.original_response = response
        self.timestamp = self._memento.timestamp

    def earlier_response(self) -> typing.Optional[typing.Any]:
        """Fetch the response earlier than this."""
        earlier_records = self._records[self._records.index(self._record) + 1:]
        response = WaybackMachineResponse(
            self.request,
            earlier_records,
            self.original_response,
            self._client
        )
        if response.is_valid():
            return response
        return None

    def is_valid(self) -> bool:
        """Whether the response is valid."""
        return self._memento is not None
